/**
  Copyright 2002 Bruce Spencer 
*/

import java.io.*;
import java.util.*;

/**
  Create a stream of tokens and delimiters from a file.
  Provides peekToken as well as nextToken.
  Delimiters are always one character Strings.
  Comment lines in files are ignored.  They are denoted as %...<eol>
  Items within single quotes are taken as one token.
  A token type is also provided, taking possible values of
  STRING, QUOTED_STRING and DELIMITER.

  White space is ignored (Tab and space characters) and
  newline characters are ignored.

  d:/MyFiles/jtp/ReadFO/TokenStream
 
  Created: Fri Jun 02 23:08:01 2000
 
  @author Bruce Spencer

 */

public class TokenStream  {

  /* An inner class for the tokens and their types */
  public class Token{
    public Token(String tokenName, int tokenType){
      this.tokenName = tokenName;
      this.tokenType = tokenType;
    }

    public Token (String tokenName){
      this(tokenName, STRING);
    }

    String tokenName;
    int tokenType;

    String getTokenName(){
      return tokenName;
    }

    int getTokenType(){
      return tokenType;
    }

    public String toString(){
      String tokenTypeName = "";
      if(tokenType == 0)
	tokenTypeName = "STRING";
      else if(tokenType == 1)
	tokenTypeName = "QUOTED_STRING";
      else if (tokenType == 2)
	tokenTypeName = "DELIMITER";
      return "Token[ tokenName = " + tokenName +
	", tokenType = " + tokenTypeName + "]";
    }
  }

  public static final int FILE_MODE = 0;
  public static final int STRING_MODE = 1;
  private int mode;

  // Token types
  public static final int STRING = 0;
  public static final int QUOTED_STRING = 1;
  public static final int DELIMITER = 2;

  private FileReader fr;
  private BufferedReader br;
  private String st;
  private StringTokenizer stt;
  private boolean peekedAhead;
  private Token token;
  private boolean atEOF;
  private String delimiters;

  /**
     Useful for error reporting.  Keeps track of the current read line.
   */
  private int lineNumber;
  public int getLineNumber () {
    return lineNumber;
  }

  /**
     @param fileName name of file to tokenize
     @param delimiters a String of characters that each act as delimiters
  */
  public TokenStream(String inString, String delimiters, int mode) throws IOException {
    this.delimiters = delimiters;
    this.mode = mode;
    if(mode == FILE_MODE){
      fr = new FileReader(inString);
      br = new BufferedReader(fr);
      st = br.readLine();
      lineNumber = 1;
      atEOF = false;
    }
    else if(mode == STRING_MODE){
      st = inString;
      lineNumber = 0;
      atEOF = true;
    }
    stt = new StringTokenizer(st, delimiters, true);
    peekedAhead = false;
  }

  /**
     @return the next non-white, non-commented token in the file.
     Delimiters are single-character strings, as defined in
     ParserConstants. A single-quote-delimited string is also a token.
     @return true is the stream has more tokens
     @see ParserContents.DELIMITERS
   */
  public boolean hasMoreTokens() throws IOException, ParseException{
    if (peekedAhead){
      return true;
    }
    else if (stt.hasMoreTokens()){
      String tempTokenName = stt.nextToken();
      if(tempTokenName.compareTo(ParserBasic.SINGLE_QUOTE)==0){
	StringBuffer sb = new StringBuffer();
	boolean quoting = true;
	while(quoting && stt.hasMoreTokens()){
	  tempTokenName = stt.nextToken();
	  if(tempTokenName.compareTo(ParserBasic.SINGLE_QUOTE)==0)
	    quoting = false;
	  else{
	    sb.append(tempTokenName);
	  }
	}
	if(quoting)
	  throw new ParseException
	    ("End of token stream within a quoted expression.");
	token = new Token(new String(sb), QUOTED_STRING);
	peekedAhead = true;
	return true;
      }
      else if(tempTokenName.equals(ParserBasic.TAB)
	      ||tempTokenName.equals(ParserBasic.SPACE)){
	//skip white space
	peekedAhead=false;
	return hasMoreTokens();
      }
      else if (tempTokenName.equals(ParserBasic.PERCENT)) {
	//throw out current stringTokenize
	//no processing of comments is done
	stt = new StringTokenizer("");
	return hasMoreTokens();
      }
      else {
	peekedAhead = true;
	int tempTokenType = STRING;
	if(ParserBasic.delimiter(tempTokenName))
	  tempTokenType = DELIMITER;
	token = new Token(tempTokenName, tempTokenType);
	return true;
      }
    }
    else if (mode == STRING_MODE) {
      return false;
    }
    //mode = FILE_MODE
    else if (atEOF) {
      return false;
    }
    else {//get the next line
      st = br.readLine();
      if (st == null) {
	fr.close();
	atEOF = true;
	return false;
      }
      else {
	lineNumber++;
	stt = new StringTokenizer(st, delimiters, true);
	return hasMoreTokens();
      }
    }
  }

  /**
     @return but do not consume the next token in the input stream
   */
  public Token peekToken() throws IOException, ParseException{
    if (hasMoreTokens())
      return token;
    else{
      throw new NoSuchElementException("No token can be found.");
    }
  }
      

  /**
     @return and consume the next token in the input stream
   */
  public Token nextToken() throws IOException, ParseException {
    if(hasMoreTokens()){
      peekedAhead = false;
      return token;
    }
    else{
      throw new NoSuchElementException("No token can be found.");
    }
  }

  public static void main(String args[]){
    try{
      TokenStream 
      t = 
	new TokenStream
	  ("top < - regular('Ho nda'). %fref", 
	   ParserBasic.DELIMITERS, STRING_MODE);
      while(t.hasMoreTokens()){
	System.out.println("Token is:" + t.nextToken());
      }
    } catch (IOException e){
      e.printStackTrace();
    } catch (ParseException e){
      e.printStackTrace();
    }
  }

} // TokenStream

  
